/* X Level - Explore a huge world */
/* Copyright (C) 1999 by David A. Madore (david.madore@ens.fr) */

/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* labyrinth.c: contains the world building routines */

#include <stdlib.h>

#include "md5.h"
#include "labyrinth.h"

unsigned long
basic_random (const char *magic_string,
	      uint64_t x, uint64_t y, uint64_t extra)
{
  struct MD5Context context;
  int i;
  unsigned char b;
  unsigned char digest[16];

  MD5Init (&context);
  MD5Update (&context, (unsigned char *)magic_string, strlen (magic_string));
  for ( i=0 ; i<8 ; i++ )
    {
      b = (x>>(i*8)) & 0xff;
      MD5Update (&context, &b, 1);
    }
  for ( i=0 ; i<8 ; i++ )
    {
      b = (y>>(i*8)) & 0xff;
      MD5Update (&context, &b, 1);
    }
  for ( i=0 ; i<8 ; i++ )
    {
      b = (extra>>(i*8)) & 0xff;
      MD5Update (&context, &b, 1);
    }
  MD5Final (digest, &context);
  return digest[0]
    + (((unsigned long) digest[1]) << 8)
    + (((unsigned long) digest[2]) << 16)
    + (((unsigned long) digest[3]) << 24);;
}

typedef struct district_requirement_s {
  unsigned char nature;
  char open_left, open_up, open_right, open_down;
} sector_t[SECTOR_SIZE][SECTOR_SIZE];

#define SECTOR_CACHE_SIZE 8

struct sector_cache_entry_s {
  char valid;
  uint64_t x;
  uint64_t y;
  sector_t d;
} sector_cache[SECTOR_CACHE_SIZE];

struct sector_cache_entry_s *
sector_new_cache_entry (void)
{
  int i;

  for ( i=0 ; i<SECTOR_CACHE_SIZE ; i++ )
    if ( ! sector_cache[i].valid )
      return &sector_cache[i];
  return &sector_cache[rand()%SECTOR_CACHE_SIZE];
}

void
make_mazelike_sector (uint64_t x, uint64_t y, sector_t d)
{
  int i, j, ii, jj, l;
  char vis[SECTOR_SIZE][SECTOR_SIZE];
  struct { int x, y; } path[SECTOR_SIZE*SECTOR_SIZE+1];
  int pathlen, count;

  for ( i=0 ; i<SECTOR_SIZE ; i++ )
    for ( j=0 ; j<SECTOR_SIZE ; j++ )
      {
	d[i][j].nature = 0;
	d[i][j].open_left = (i==0) && (j==SECTOR_SIZE/2);
	d[i][j].open_up = (j==0) && (i==SECTOR_SIZE/2);
	d[i][j].open_right = (i==SECTOR_SIZE-1) && (j==SECTOR_SIZE/2);
	d[i][j].open_down = (j==SECTOR_SIZE-1) && (i==SECTOR_SIZE/2);
	vis[i][j] = 0;
      }
  pathlen = 1;  path[0].x = SECTOR_SIZE/2;  path[0].y = SECTOR_SIZE/2;
  vis[path[0].x][path[0].y] = 1;
  count = 0;
  while ( 1 )
    {
      while ( pathlen )
	{
	  i = path[pathlen-1].x;  j = path[pathlen-1].y;
	  if ( ( i > 0 && ! vis[i-1][j] )
	       || ( j > 0 && ! vis[i][j-1] )
	       || ( i < SECTOR_SIZE-1 && ! vis[i+1][j] )
	       || ( j < SECTOR_SIZE-1 && ! vis[i][j+1] ) )
	    break;
	  pathlen--;
	}
      if ( ! pathlen )
	break;
      do
	{
	  switch ( l = (basic_random ("The Doors of Twilight",
				      x, y, count++) % 4) )
	    {
	    case 0: ii = i-1; jj = j; break;
	    case 1: ii = i; jj = j-1; break;
	    case 2: ii = i+1; jj = j; break;
	    case 3: ii = i; jj = j+1; break;
	    }
	}
      while ( ii < 0 || jj < 0 || ii >= SECTOR_SIZE || jj >= SECTOR_SIZE
	      || vis[ii][jj] );
      path[pathlen].x = ii;  path[pathlen].y = jj;
      pathlen++;
      vis[ii][jj] = 1;
      switch ( l )
	{
	case 0: d[i][j].open_left = 1;  d[ii][jj].open_right = 1; break;
	case 1: d[i][j].open_up = 1;  d[ii][jj].open_down = 1; break;
	case 2: d[i][j].open_right = 1;  d[ii][jj].open_left = 1; break;
	case 3: d[i][j].open_down = 1;  d[ii][jj].open_up = 1; break;
	}
    }
  count = 0;
  while ( basic_random ("To make a trap for fools", x, y, count++) % 360 )
    {
      i = basic_random ("The moving finger writes",
			x, y, count) % SECTOR_SIZE;
      j = basic_random ("And having writ moves on",
			x, y, count) % SECTOR_SIZE;
      if ( i > 0 )
	{
	  d[i][j].open_left = 1;
	  d[i-1][j].open_right = 1;
	}
      if ( j > 0 )
	{
	  d[i][j].open_up = 1;
	  d[i][j-1].open_down = 1;
	}
      if ( i < SECTOR_SIZE - 1 )
	{
	  d[i][j].open_right = 1;
	  d[i+1][j].open_left = 1;
	}
      if ( j < SECTOR_SIZE - 1 )
	{
	  d[i][j].open_down = 1;
	  d[i][j+1].open_up = 1;
	}
    }
  count = 0;
  while ( basic_random ("The Jewel in the Crown", x, y, count++) % 250 )
    {
      i = basic_random ("Eternity is long",
			x, y, count) % SECTOR_SIZE;
      j = basic_random ("Especially near the end",
			x, y, count) % SECTOR_SIZE;
      d[i][j].nature = 1;
    }
  count = 0;
  while ( basic_random ("Errare humanum est", x, y, count++) % 800 )
    {
      i = basic_random ("De omni re scibili",
			x, y, count) % SECTOR_SIZE;
      j = basic_random ("Et quibusdam aliis",
			x, y, count) % SECTOR_SIZE;
      if ( d[i][j].open_left + d[i][j].open_up
	   + d[i][j].open_right + d[i][j].open_down == 1 )
	{
	  if ( basic_random ("Non erat his locus", x, y, count) % 4 )
	    d[i][j].nature = 1;
	  if ( d[i][j].open_left )
	    {
	      d[i][j].open_left = 0;
	      d[i-1][j].open_right = 0;
	    }
	  if ( d[i][j].open_up )
	    {
	      d[i][j].open_up = 0;
	      d[i][j-1].open_down = 0;
	    }
	  if ( d[i][j].open_right )
	    {
	      d[i][j].open_right = 0;
	      d[i+1][j].open_left = 0;
	    }
	  if ( d[i][j].open_down )
	    {
	      d[i][j].open_down = 0;
	      d[i][j+1].open_up = 0;
	    }
	}
    }
}

void
make_trans_sector (uint64_t x, uint64_t y, sector_t d)
{
  int n, i, j, ii, jj, l;
  int count;

  for ( i=0 ; i<SECTOR_SIZE ; i++ )
    for ( j=0 ; j<SECTOR_SIZE ; j++ )
      {
	d[i][j].nature = 1;
	d[i][j].open_left = (i==0) && (j==SECTOR_SIZE/2);
	d[i][j].open_up = (j==0) && (i==SECTOR_SIZE/2);
	d[i][j].open_right = (i==SECTOR_SIZE-1) && (j==SECTOR_SIZE/2);
	d[i][j].open_down = (j==SECTOR_SIZE-1) && (i==SECTOR_SIZE/2);
      }
  count = 0;
  for ( n=0 ; n<4 ; n++ )
    {
      i = SECTOR_SIZE/2;  j = SECTOR_SIZE/2;
      switch ( n )
	{
	case 0: ii = 0;  jj = SECTOR_SIZE/2; break;
	case 1: ii = SECTOR_SIZE/2;  jj = 0; break;
	case 2: ii = SECTOR_SIZE-1;  jj = SECTOR_SIZE/2; break;
	case 3: ii = SECTOR_SIZE/2;  jj = SECTOR_SIZE-1; break;
	}
      while ( i != ii || j != jj )
	{
	  if ( (basic_random ("Open your eyes and open your mind",
			      x+i, y+j, count) % 15) == 0 )
	    d[i][j].nature = 0;
	  l = basic_random ("Seek the answers in your heart",
			    x, y, count++) % 6;
	  if ( l == 4 )
	    l = n;
	  else if ( l == 5 )
	    {
	      if ( abs (i-ii) > abs (j-jj) )
		{
		  if ( i > ii )
		    l = 0;
		  else
		    l = 2;
		}
	      else
		{
		  if ( j > jj )
		    l = 1;
		  else
		    l = 3;
		}
	    }
	  switch ( l )
	    {
	    case 0:
	      if ( i > 0 )
		{
		  d[i][j].open_left = 1;
		  i--;
		  d[i][j].open_right = 1;
		}
	      break;
	    case 1:
	      if ( j > 0 )
		{
		  d[i][j].open_up = 1;
		  j--;
		  d[i][j].open_down = 1;
		}
	      break;
	    case 2:
	      if ( i < SECTOR_SIZE-1 )
		{
		  d[i][j].open_right = 1;
		  i++;
		  d[i][j].open_left = 1;
		}
	      break;
	    case 3:
	      if ( j < SECTOR_SIZE-1 )
		{
		  d[i][j].open_down = 1;
		  j++;
		  d[i][j].open_up = 1;
		}
	      break;
	    }
	}
    }
}

sector_t *
scratch_read_sector (uint64_t x, uint64_t y)
{
  struct sector_cache_entry_s *ce;
  sector_t *d;
  unsigned char b;

  ce = sector_new_cache_entry ();
  ce->valid = 1;
  ce->x = x;
  ce->y = y;
  d = &ce->d;
  b = basic_random ("The Halls of Montezuma", x, y, 0);
  if ( b%4 == 2 )
    make_mazelike_sector (x, y, *d);
  else
    make_trans_sector (x, y, *d);
  return d;
}

sector_t *
read_sector (uint64_t x, uint64_t y)
{
  int i;

  for ( i=0 ; i<SECTOR_CACHE_SIZE ; i++ )
    if ( sector_cache[i].valid
	 && sector_cache[i].x == x && sector_cache[i].y == y )
      return &sector_cache[i].d;
  return scratch_read_sector (x, y);
}

typedef unsigned char district_t[DISTRICT_SIZE][DISTRICT_SIZE];

#define DISTRICT_CACHE_SIZE 32

struct district_cache_entry_s {
  char valid;
  uint64_t x;
  uint64_t y;
  district_t d;
} district_cache[DISTRICT_CACHE_SIZE];

struct district_cache_entry_s *
district_new_cache_entry (void)
{
  int i;

  for ( i=0 ; i<DISTRICT_CACHE_SIZE ; i++ )
    if ( ! district_cache[i].valid )
      return &district_cache[i];
  return &district_cache[rand()%DISTRICT_CACHE_SIZE];
}

void
make_block_district (uint64_t x, uint64_t y, district_t d,
		     struct district_requirement_s rq)
{
  int i, j;
  unsigned char b;

  b = basic_random ("The Shores of Tripoli", x, y, 0);
  for ( i=0 ; i<DISTRICT_SIZE ; i++ )
    for ( j=0 ; j<DISTRICT_SIZE ; j++ )
      d[i][j] = 1;
  if ( rq.open_left )
    for ( i=0 ; i<=DISTRICT_SIZE/2 ; i++ )
      {
	d[i][DISTRICT_SIZE/2] = 0;
	d[i][DISTRICT_SIZE/2-1] = 0;
      }
  if ( rq.open_up )
    for ( i=0 ; i<=DISTRICT_SIZE/2 ; i++ )
      {
	d[DISTRICT_SIZE/2][i] = 0;
	d[DISTRICT_SIZE/2-1][i] = 0;
      }
  if ( rq.open_right )
    for ( i=DISTRICT_SIZE/2-1 ; i<DISTRICT_SIZE ; i++ )
      {
	d[i][DISTRICT_SIZE/2] = 0;
	d[i][DISTRICT_SIZE/2-1] = 0;
      }
  if ( rq.open_down )
    for ( i=DISTRICT_SIZE/2-1 ; i<DISTRICT_SIZE ; i++ )
      {
	d[DISTRICT_SIZE/2][i] = 0;
	d[DISTRICT_SIZE/2-1][i] = 0;
      }
  if ( rq.open_left+rq.open_up+rq.open_right+rq.open_down >= 3
       && b%2 )
    {
      for ( i=DISTRICT_SIZE/2-2 ; i<DISTRICT_SIZE/2+2 ; i++ )
	for ( j=DISTRICT_SIZE/2-2 ; j<DISTRICT_SIZE/2+2 ; j++ )
	  d[i][j] = 0;
    }
}

void
make_room_district (uint64_t x, uint64_t y, district_t d,
		     struct district_requirement_s rq)
{
  int i, j;
  unsigned char b;

  b = basic_random ("The Shores of Tripoli", x, y, 0);
  for ( i=0 ; i<DISTRICT_SIZE ; i++ )
    for ( j=0 ; j<DISTRICT_SIZE ; j++ )
      d[i][j] = ( i==0 || j==0
		  || i==DISTRICT_SIZE-1 || j==DISTRICT_SIZE-1 );
  if ( rq.open_left ) d[0][DISTRICT_SIZE/2] = 0;
  if ( rq.open_left ) d[0][DISTRICT_SIZE/2-1] = 0;
  if ( rq.open_up ) d[DISTRICT_SIZE/2][0] = 0;
  if ( rq.open_up ) d[DISTRICT_SIZE/2-1][0] = 0;
  if ( rq.open_right ) d[DISTRICT_SIZE-1][DISTRICT_SIZE/2] = 0;
  if ( rq.open_right ) d[DISTRICT_SIZE-1][DISTRICT_SIZE/2-1] = 0;
  if ( rq.open_down ) d[DISTRICT_SIZE/2][DISTRICT_SIZE-1] = 0;
  if ( rq.open_down ) d[DISTRICT_SIZE/2-1][DISTRICT_SIZE-1] = 0;
  if ( b%5 == 0 )
    for ( i=DISTRICT_SIZE/4 ; i<3*DISTRICT_SIZE/4 ; i++ )
      for ( j=DISTRICT_SIZE/4 ; j<3*DISTRICT_SIZE/4 ; j++ )
	d[i][j] = 1;
  if ( b%7 == 0 )
    {
      for ( i=4 ; i<DISTRICT_SIZE-4 ; i++ )
	d[i][2] = 1;
      for ( i=4 ; i<DISTRICT_SIZE-4 ; i++ )
	d[i][DISTRICT_SIZE-3] = 1;
      for ( i=4 ; i<DISTRICT_SIZE-4 ; i++ )
	d[2][i] = 1;
      for ( i=4 ; i<DISTRICT_SIZE-4 ; i++ )
	d[DISTRICT_SIZE-3][i] = 1;
    }
  if ( b%9 == 0 )
    {
      if ( b & 64 )
	for ( i=3 ; i<DISTRICT_SIZE-3 ; i++ )
	  {
	    d[DISTRICT_SIZE/2][i] = 1;
	    d[DISTRICT_SIZE/2-1][i] = 1;
	  }
      if ( b & 128 )
	for ( i=3 ; i<DISTRICT_SIZE-3 ; i++ )
	  {
	    d[i][DISTRICT_SIZE/2] = 1;
	    d[i][DISTRICT_SIZE/2-1] = 1;
	  }
    }
}

district_t *
scratch_read_district (uint64_t x, uint64_t y)
{
  struct district_cache_entry_s *ce;
  district_t *d;
  struct district_requirement_s rq;

  ce = district_new_cache_entry ();
  ce->valid = 1;
  ce->x = x;
  ce->y = y;
  d = &ce->d;
  rq = ((*(read_sector ((x/SECTOR_FULL_SIZE)*SECTOR_FULL_SIZE,
		       (y/SECTOR_FULL_SIZE)*SECTOR_FULL_SIZE)))
	[(x/DISTRICT_SIZE)%SECTOR_SIZE][(y/DISTRICT_SIZE)%SECTOR_SIZE]);
  if ( rq.nature )
    make_block_district (x, y, *d, rq);
  else
    make_room_district (x, y, *d, rq);
  return d;
}

district_t *
read_district (uint64_t x, uint64_t y)
{
  int i;

  for ( i=0 ; i<DISTRICT_CACHE_SIZE ; i++ )
    if ( district_cache[i].valid
	 && district_cache[i].x == x && district_cache[i].y == y )
      return &district_cache[i].d;
  return scratch_read_district (x, y);
}

unsigned char
read_square (uint64_t x, uint64_t y)
{
  district_t *d;

  d = read_district ((x/DISTRICT_SIZE)*DISTRICT_SIZE,
		     (y/DISTRICT_SIZE)*DISTRICT_SIZE);
  return (*d)[x%DISTRICT_SIZE][y%DISTRICT_SIZE];
}
