/* X Level - Explore a huge world */
/* Copyright (C) 1999 by David A. Madore (david.madore@ens.fr) */

/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* General include files */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#ifdef HAVE_STDINT_H
#include <stdint.h>
#endif
#include <time.h>
#include <pwd.h>
#include <sys/types.h>
#include <unistd.h>

/* Xlib include files */
#include <X11/Xlib.h>

/* Xt include files */
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>

/* Xaw include files */
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Toggle.h>

#include "labyrinth.h"

long int hero_dx, hero_dy;
uint64_t hero_x, hero_y;
uint64_t hero_oldx, hero_oldy;
uint64_t corner_x, corner_y;

char hero_moved_p;
char recentered_p;

#define APP_CLASS "XLevel"

#define WIDTH_S 32
#define HEIGHT_S 32

#define SQWIDTH 12
#define SQHEIGHT 12

#define WIDTH_P (WIDTH_S*SQWIDTH)
#define HEIGHT_P (HEIGHT_S*SQHEIGHT)

#define BASETIME 50

String fallback_resources[] = {
  "*input: True",
  "*background: Grey51",
  "*foreground: White",
  "*gameSpace.background: MidnightBlue",
  "*Box.background: DarkSlateGrey",
  "*Label.foreground: White",
  "*label.foreground: Green",
  "*Command.foreground: Yellow",
  "*quitButton.label: Quit",
  "*quitButton.accelerators:"
  "<KeyPress>Escape: set() notify()",
  "*leftButton.label: L",
  "*leftButton.accelerators:"
  "<KeyPress>Left: set() notify() unset()\\n\\\n"
  "<KeyPress>KP_Left: set() notify() unset()",
  "*upButton.label: U",
  "*upButton.accelerators:"
  "<KeyPress>Up: set() notify() unset()\\n\\\n"
  "<KeyPress>KP_Up: set() notify() unset()",
  "*rightButton.label: R",
  "*rightButton.accelerators:"
  "<KeyPress>Right: set() notify() unset()\\n\\\n"
  "<KeyPress>KP_Right: set() notify() unset()",
  "*downButton.label: D",
  "*downButton.accelerators:"
  "<KeyPress>Down: set() notify() unset()\\n\\\n"
  "<KeyPress>KP_Down: set() notify() unset()",
  "*stopButton.label: D",
  "*stopButton.accelerators:"
  "<KeyPress>KP_Begin: set() notify() unset()",
  NULL
};

typedef struct app_res_val_s {
  Pixel color1, color2, color3, color4;
} app_res_val_t;

app_res_val_t app_res_val;

XtResource app_res_tab[] = {
  {"color1", "Color", XtRPixel, sizeof(Pixel),
   XtOffsetOf (app_res_val_t, color1),
   XtRString, "IndianRed"
  },
  {"color2", "Color", XtRPixel, sizeof(Pixel),
   XtOffsetOf (app_res_val_t, color2),
   XtRString, "GoldenRod"
  },
  {"color3", "Color", XtRPixel, sizeof(Pixel),
   XtOffsetOf (app_res_val_t, color3),
   XtRString, "SeaGreen"
  },
  {"color4", "Color", XtRPixel, sizeof(Pixel),
   XtOffsetOf (app_res_val_t, color4),
   XtRString, "Beige"
  },
};

XtAppContext app_context;
XtIntervalId timeout;

Widget toplevel;
Widget ground_box;
Widget label;
Widget quit_button;
Widget left_button, up_button, right_button, down_button, stop_button;
Widget game_space;

Display *dpy;
Window win;
GC gc, gcb, gc1, gc2, gc3, gc4;
Pixmap hero_pm;
Pixmap pm[3];

void
update_game (void)
{
  hero_oldx = hero_x;
  hero_oldy = hero_y;
  switch ( read_square (hero_x+hero_dx, hero_y+hero_dy) )
    {
    case 0:
      hero_x += hero_dx;
      hero_y += hero_dy;
      hero_moved_p = ( hero_dx || hero_dy );
      break;
    case 1:
      hero_dx = 0;
      hero_dy = 0;
      hero_moved_p = 0;
      break;
    case 2:
      hero_dx = -hero_dx;
      hero_dy = -hero_dy;
      hero_moved_p = 0;
      break;
    }
  recentered_p = 0;
  while ( hero_x-corner_x < WIDTH_S/4 )
    {
      corner_x --;
      recentered_p = 1;
    }
  while ( hero_y-corner_y < HEIGHT_S/4 )
    {
      corner_y --;
      recentered_p = 1;
    }
  while ( hero_x-corner_x >= 3*WIDTH_S/4 )
    {
      corner_x ++;
      recentered_p = 1;
    }
  while ( hero_y-corner_y >= 3*HEIGHT_S/4 )
    {
      corner_y ++;
      recentered_p = 1;
    }
}

void
draw_square (int rx, int ry)
{
  unsigned char v;

  if ( rx < 0 || ry < 0 || rx >= WIDTH_S || ry >= HEIGHT_S )
    return;
  v = read_square (rx+corner_x, ry+corner_y);
  if ( v >= 0 && v <= 3 )
    XCopyArea (dpy, pm[v], win, gc,
	       0, 0, SQWIDTH, SQHEIGHT,
	       rx*SQWIDTH, ry*SQHEIGHT);
  if ( rx+corner_x == hero_x && ry+corner_y == hero_y )
    XCopyArea (dpy, hero_pm, win, gc,
	       0, 0, SQWIDTH, SQHEIGHT,
	       rx*SQWIDTH, ry*SQHEIGHT);
}

void
draw_game (char redraw_all)
{
  int i, j;

  if ( redraw_all || recentered_p )
    for ( i=0 ; i<WIDTH_S ; i++ )
      for ( j=0 ; j<HEIGHT_S ; j++ )
	draw_square (i, j);
  else if ( hero_moved_p )
    {
      draw_square (hero_x - corner_x, hero_y - corner_y);
      draw_square (hero_oldx - corner_x, hero_oldy - corner_y);
    }
}

void
refresh (void)
{
  XFillRectangle (dpy, win, gcb, 0, 0, WIDTH_P, HEIGHT_P);
  draw_game (1);
}

void
hero_dir (int dx, int dy)
{
  hero_dx = dx;
  hero_dy = dy;
}

void
left_callback (Widget w, XtPointer client_data, XtPointer call_data)
{
  hero_dir (-1, 0);
}

void
up_callback (Widget w, XtPointer client_data, XtPointer call_data)
{
  hero_dir (0, -1);
}

void
right_callback (Widget w, XtPointer client_data, XtPointer call_data)
{
  hero_dir (1, 0);
}

void
down_callback (Widget w, XtPointer client_data, XtPointer call_data)
{
  hero_dir (0, 1);
}

void
stop_callback (Widget w, XtPointer client_data, XtPointer call_data)
{
  hero_dir (0, 0);
}

void
quit_callback (Widget w, XtPointer client_data, XtPointer call_data)
     /* Callback called by the quit button */
{
  exit (0);
}

void
expose_proc (Widget w, XtPointer client_data, XEvent *evt, Boolean *cnt)
     /* Event handler called when an Expose event is received */
{
  refresh ();
  *cnt = False;
}

void timer_proc (XtPointer client_data, XtIntervalId *id);

void
retime (void)
     /* Make sure the game is running. */
{
  timeout = XtAppAddTimeOut (app_context, BASETIME, timer_proc, NULL);
}

void
timer_proc (XtPointer client_data, XtIntervalId *id)
     /* Timeout handler called every time unit */
{
  update_game ();
  draw_game (0);
  retime ();
}

int
main (int argc, char *argv[])
{
  toplevel = XtVaAppInitialize(&app_context, APP_CLASS,
			       NULL, 0,
			       &argc, argv, fallback_resources,
			       NULL);
  XtGetApplicationResources (toplevel, &app_res_val, app_res_tab,
			     XtNumber (app_res_tab), NULL, 0);
  ground_box =
    XtVaCreateManagedWidget ("groundBox", boxWidgetClass, toplevel,
			     NULL);
  label =
    XtVaCreateManagedWidget ("label", labelWidgetClass, ground_box,
			     XtNlabel, "xlevel",
			     NULL);
  quit_button =
    XtVaCreateManagedWidget ("quitButton", commandWidgetClass, ground_box,
			     NULL);
  left_button =
    XtVaCreateManagedWidget ("leftButton", commandWidgetClass, ground_box,
			     NULL);
  up_button =
    XtVaCreateManagedWidget ("upButton", commandWidgetClass, ground_box,
			     NULL);
  right_button =
    XtVaCreateManagedWidget ("rightButton", commandWidgetClass, ground_box,
			     NULL);
  down_button =
    XtVaCreateManagedWidget ("downButton", commandWidgetClass, ground_box,
			     NULL);
  stop_button =
    XtVaCreateManagedWidget ("stopButton", commandWidgetClass, ground_box,
			     NULL);
  game_space =
    XtVaCreateManagedWidget ("gameSpace", coreWidgetClass, ground_box,
			     XtNwidth, WIDTH_P,
			     XtNheight, HEIGHT_P,
			     NULL);
  XtRealizeWidget(toplevel);
  dpy = XtDisplay (game_space);
  win = XtWindow (game_space);
  gc = XCreateGC (dpy, win, 0, NULL);
  gcb = XCreateGC (dpy, win, 0, NULL);
  gc1 = XCreateGC (dpy, win, 0, NULL);
  gc2 = XCreateGC (dpy, win, 0, NULL);
  gc3 = XCreateGC (dpy, win, 0, NULL);
  gc4 = XCreateGC (dpy, win, 0, NULL);
  {
    Pixel bgc;
    XtVaGetValues (game_space,
		   XtNbackground, &bgc,
		   NULL);
    XSetBackground (dpy, gc1, bgc);
    XSetBackground (dpy, gc2, bgc);
    XSetBackground (dpy, gc3, bgc);
    XSetBackground (dpy, gc4, bgc);
    XSetForeground (dpy, gcb, bgc);
    XSetForeground (dpy, gc1, app_res_val.color1);
    XSetForeground (dpy, gc2, app_res_val.color2);
    XSetForeground (dpy, gc3, app_res_val.color3);
    XSetForeground (dpy, gc4, app_res_val.color4);
  }
  {
    hero_pm = XCreatePixmap (dpy, win, SQWIDTH, SQHEIGHT,
			     DefaultDepth (dpy, DefaultScreen (dpy)));
    XFillRectangle (dpy, hero_pm, gcb, 0, 0, SQWIDTH, SQHEIGHT);
    XFillArc (dpy, hero_pm, gc1, 0, 0, SQWIDTH, SQHEIGHT, 0, 23040);
    XDrawPoint (dpy, hero_pm, gc2, SQWIDTH/4, 3*SQHEIGHT/8);
    XDrawPoint (dpy, hero_pm, gc2, 3*SQWIDTH/4, 3*SQHEIGHT/8);
    XDrawArc (dpy, hero_pm, gc2, SQWIDTH/4, 3*SQHEIGHT/8,
	      SQWIDTH/2, 3*SQHEIGHT/8, 11520, 12000);
    pm[0] = XCreatePixmap (dpy, win, SQWIDTH, SQHEIGHT,
			   DefaultDepth (dpy, DefaultScreen (dpy)));
    XFillRectangle (dpy, pm[0], gcb, 0, 0, SQWIDTH, SQHEIGHT);
    pm[1] = XCreatePixmap (dpy, win, SQWIDTH, SQHEIGHT,
			   DefaultDepth (dpy, DefaultScreen (dpy)));
    XFillRectangle (dpy, pm[1], gcb, 0, 0, SQWIDTH, SQHEIGHT);
    XFillRectangle (dpy, pm[1], gc4, 0, 0, SQWIDTH, SQHEIGHT);
    pm[2] = XCreatePixmap (dpy, win, SQWIDTH, SQHEIGHT,
			   DefaultDepth (dpy, DefaultScreen (dpy)));
    XFillRectangle (dpy, pm[2], gcb, 0, 0, SQWIDTH, SQHEIGHT);
    XFillRectangle (dpy, pm[2], gc3, 0, 0, SQWIDTH, SQHEIGHT);
  }
  XtAddCallback (quit_button, XtNcallback, quit_callback, NULL);
  XtAddCallback (left_button, XtNcallback, left_callback, NULL);
  XtAddCallback (up_button, XtNcallback, up_callback, NULL);
  XtAddCallback (right_button, XtNcallback, right_callback, NULL);
  XtAddCallback (down_button, XtNcallback, down_callback, NULL);
  XtAddCallback (stop_button, XtNcallback, stop_callback, NULL);
  XtAddEventHandler (game_space, ExposureMask, False, expose_proc, NULL);
  XtInstallAllAccelerators (game_space, ground_box);
  XtInstallAllAccelerators (ground_box, ground_box);
  hero_x = DISTRICT_SIZE/2;
  hero_y = DISTRICT_SIZE/2;
  corner_x = hero_x-WIDTH_S/2;
  corner_y = hero_y-HEIGHT_S/2;
  hero_dir (0, 0);
  retime ();
  XtAppMainLoop (app_context);
  return 0;
}
